@@ -17,7 +17,7 @@
 %                               December 2003                                 %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2021 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999 ImageMagick Studio LLC, a non-profit organization           %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -44,6 +44,7 @@
 #include "wand/studio.h"
 #include "wand/MagickWand.h"
 #include "wand/mogrify-private.h"
+#include "magick/image-private.h"
 #include "magick/string-private.h"
 
 /*
@@ -194,7 +195,7 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
   int argc,char **argv,char **metadata,ExceptionInfo *exception)
 {
 #define CompareEpsilon  (1.0e-06)
-#define DefaultDissimilarityThreshold  0.31830988618379067154
+#define DefaultDissimilarityThreshold  (1.0/MagickPI)
 #define DefaultSimilarityThreshold  (-1.0)
 #define DestroyCompare() \
 { \
@@ -241,7 +242,7 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
 
   Image
     *difference_image,
-    *image,
+    *image = (Image *) NULL,
     *reconstruct_image,
     *similarity_image;
 
@@ -255,6 +256,7 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
     fire,
     pend,
     respect_parenthesis,
+    similar = MagickTrue,
     subimage_search;
 
   MagickStatusType
@@ -292,7 +294,12 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
         }
     }
   if (argc < 3)
-    return(CompareUsage());
+    {
+      (void) ThrowMagickException(exception,GetMagickModule(),OptionError,
+        "MissingArgument","%s","");
+      (void) CompareUsage();
+      return(MagickFalse);
+    }
   restore_info=image_info;
   channels=DefaultChannels;
   difference_image=NewImageList();
@@ -1159,69 +1166,82 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
       similarity_image=SimilarityMetricImage(image,reconstruct_image,metric,
         &offset,&similarity_metric,exception);
       if (similarity_metric > dissimilarity_threshold)
-        ThrowCompareException(ImageError,"ImagesTooDissimilar",image->filename);
+        (void) ThrowMagickException(exception,GetMagickModule(),ImageWarning,
+          "ImagesTooDissimilar","`%s'",image->filename);
     }
-  if ((reconstruct_image->columns == image->columns) &&
-       (reconstruct_image->rows == image->rows))
+  if (similarity_image == (Image *) NULL)
     difference_image=CompareImageChannels(image,reconstruct_image,channels,
       metric,&distortion,exception);
   else
-    if (similarity_image == (Image *) NULL)
-      difference_image=CompareImageChannels(image,reconstruct_image,channels,
-        metric,&distortion,exception);
-    else
-      {
-        Image
-          *composite_image;
+    {
+      Image
+        *composite_image;
 
-        /*
-          Determine if reconstructed image is a subimage of the image.
-        */
-        composite_image=CloneImage(image,0,0,MagickTrue,exception);
-        if (composite_image == (Image *) NULL)
-          difference_image=CompareImageChannels(image,reconstruct_image,
+      /*
+        Determine if reconstructed image is a subimage of the image.
+      */
+      composite_image=CloneImage(image,0,0,MagickTrue,exception);
+      if (composite_image == (Image *) NULL)
+        difference_image=CompareImageChannels(image,reconstruct_image,
+          channels,metric,&distortion,exception);
+      else
+        {
+          Image
+            *distort_image;
+
+          RectangleInfo
+            page;
+
+          (void) CompositeImage(composite_image,CopyCompositeOp,
+            reconstruct_image,offset.x,offset.y);
+          difference_image=CompareImageChannels(image,composite_image,
             channels,metric,&distortion,exception);
-        else
-          {
-            Image
-              *distort_image;
+          if (difference_image != (Image *) NULL)
+            {
+              difference_image->page.x=offset.x;
+              difference_image->page.y=offset.y;
+            }
+          composite_image=DestroyImage(composite_image);
+          page.width=reconstruct_image->columns;
+          page.height=reconstruct_image->rows;
+          page.x=offset.x;
+          page.y=offset.y;
+          distort_image=CropImage(image,&page,exception);
+          if (distort_image != (Image *) NULL)
+            {
+              Image
+                *sans_image;
 
-            RectangleInfo
-              page;
-
-            (void) CompositeImage(composite_image,CopyCompositeOp,
-              reconstruct_image,offset.x,offset.y);
-            difference_image=CompareImageChannels(image,composite_image,
-              channels,metric,&distortion,exception);
-            if (difference_image != (Image *) NULL)
-              {
-                difference_image->page.x=offset.x;
-                difference_image->page.y=offset.y;
-              }
-            composite_image=DestroyImage(composite_image);
-            page.width=reconstruct_image->columns;
-            page.height=reconstruct_image->rows;
-            page.x=offset.x;
-            page.y=offset.y;
-            distort_image=CropImage(image,&page,exception);
-            if (distort_image != (Image *) NULL)
-              {
-                Image
-                  *sans_image;
-
-                sans_image=CompareImageChannels(distort_image,reconstruct_image,
-                  channels,metric,&distortion,exception);
-                distort_image=DestroyImage(distort_image);
-                if (sans_image != (Image *) NULL)
-                  sans_image=DestroyImage(sans_image);
-              }
-          }
-        if (difference_image != (Image *) NULL)
-          {
-            AppendImageToList(&difference_image,similarity_image);
-            similarity_image=(Image *) NULL;
-          }
-      }
+              sans_image=CompareImageChannels(distort_image,reconstruct_image,
+                channels,metric,&distortion,exception);
+              distort_image=DestroyImage(distort_image);
+              if (sans_image != (Image *) NULL)
+                sans_image=DestroyImage(sans_image);
+            }
+        }
+      if (difference_image != (Image *) NULL)
+        {
+          AppendImageToList(&difference_image,similarity_image);
+          similarity_image=(Image *) NULL;
+        }
+    }
+  if (fabs(distortion) > CompareEpsilon)
+    similar=MagickFalse;
+  switch (metric)
+  {
+    case NormalizedCrossCorrelationErrorMetric:
+    {
+      distortion=1.0-distortion;
+      similarity_metric=1.0-similarity_metric;
+    }
+    case PeakSignalToNoiseRatioMetric:
+    {
+      distortion=fabs(distortion);
+      similarity_metric+=1.0;
+    }
+    default:
+      break;
+  }
   if (difference_image == (Image *) NULL)
     status=0;
   else
@@ -1240,23 +1260,19 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
             case MeanSquaredErrorMetric:
             case PeakAbsoluteErrorMetric:
             case RootMeanSquaredErrorMetric:
+            case AbsoluteErrorMetric:
+            case NormalizedCrossCorrelationErrorMetric:
+            case PerceptualHashErrorMetric:
             {
               (void) FormatLocaleFile(stderr,"%.*g (%.*g)",GetMagickPrecision(),
-                QuantumRange*distortion,GetMagickPrecision(),distortion);
+                (double) QuantumRange*distortion,GetMagickPrecision(),
+                distortion);
               break;
             }
             case PeakSignalToNoiseRatioMetric:
             {
               (void) FormatLocaleFile(stderr,"%.*g (%.*g)",GetMagickPrecision(),
-                distortion,GetMagickPrecision(),0.01*distortion);
-              break;
-            }
-            case AbsoluteErrorMetric:
-            case NormalizedCrossCorrelationErrorMetric:
-            case PerceptualHashErrorMetric:
-            {
-              (void) FormatLocaleFile(stderr,"%.*g",GetMagickPrecision(),
-                distortion);
+                48.1647*distortion,GetMagickPrecision(),distortion);
               break;
             }
             case MeanErrorPerPixelMetric:
@@ -1271,8 +1287,9 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
               break;
           }
           if (subimage_search != MagickFalse)
-            (void) FormatLocaleFile(stderr," @ %.20g,%.20g",(double)
-              difference_image->page.x,(double) difference_image->page.y);
+            (void) FormatLocaleFile(stderr," @ %.20g,%.20g [%.*g]",
+              (double) offset.x,(double) offset.y,GetMagickPrecision(),
+              similarity_metric);
         }
       else
         {
@@ -1302,20 +1319,20 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
                 default:
                 {
                   (void) FormatLocaleFile(stderr,"    red: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[RedChannel],GetMagickPrecision(),
                     channel_distortion[RedChannel]);
                   (void) FormatLocaleFile(stderr,"    green: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[GreenChannel],GetMagickPrecision(),
                     channel_distortion[GreenChannel]);
                   (void) FormatLocaleFile(stderr,"    blue: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[BlueChannel],GetMagickPrecision(),
                     channel_distortion[BlueChannel]);
                   if (image->matte != MagickFalse)
                     (void) FormatLocaleFile(stderr,"    alpha: %.*g (%.*g)\n",
-                      GetMagickPrecision(),QuantumRange*
+                      GetMagickPrecision(),(double) QuantumRange*
                       channel_distortion[OpacityChannel],GetMagickPrecision(),
                       channel_distortion[OpacityChannel]);
                   break;
@@ -1323,24 +1340,24 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
                 case CMYKColorspace:
                 {
                   (void) FormatLocaleFile(stderr,"    cyan: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[CyanChannel],GetMagickPrecision(),
                     channel_distortion[CyanChannel]);
                   (void) FormatLocaleFile(stderr,"    magenta: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[MagentaChannel],GetMagickPrecision(),
                     channel_distortion[MagentaChannel]);
                   (void) FormatLocaleFile(stderr,"    yellow: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[YellowChannel],GetMagickPrecision(),
                     channel_distortion[YellowChannel]);
                   (void) FormatLocaleFile(stderr,"    black: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[BlackChannel],GetMagickPrecision(),
                     channel_distortion[BlackChannel]);
                   if (image->matte != MagickFalse)
                     (void) FormatLocaleFile(stderr,"    alpha: %.*g (%.*g)\n",
-                      GetMagickPrecision(),QuantumRange*
+                      GetMagickPrecision(),(double) QuantumRange*
                       channel_distortion[OpacityChannel],GetMagickPrecision(),
                       channel_distortion[OpacityChannel]);
                   break;
@@ -1349,19 +1366,19 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
                 case GRAYColorspace:
                 {
                   (void) FormatLocaleFile(stderr,"    gray: %.*g (%.*g)\n",
-                    GetMagickPrecision(),QuantumRange*
+                    GetMagickPrecision(),(double) QuantumRange*
                     channel_distortion[GrayChannel],GetMagickPrecision(),
                     channel_distortion[GrayChannel]);
                   if (image->matte != MagickFalse)
                     (void) FormatLocaleFile(stderr,"    alpha: %.*g (%.*g)\n",
-                      GetMagickPrecision(),QuantumRange*
+                      GetMagickPrecision(),(double) QuantumRange*
                       channel_distortion[OpacityChannel],GetMagickPrecision(),
                       channel_distortion[OpacityChannel]);
                   break;
                 }
               }
               (void) FormatLocaleFile(stderr,"    all: %.*g (%.*g)\n",
-                GetMagickPrecision(),QuantumRange*
+                GetMagickPrecision(),(double) QuantumRange*
                 channel_distortion[CompositeChannels],GetMagickPrecision(),
                 channel_distortion[CompositeChannels]);
               break;
@@ -1431,8 +1448,13 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
           channel_distortion=(double *) RelinquishMagickMemory(
             channel_distortion);
           if (subimage_search != MagickFalse)
-            (void) FormatLocaleFile(stderr,"    Offset: %.20g,%.20g\n",(double)
-              difference_image->page.x,(double) difference_image->page.y);
+            {
+              (void) FormatLocaleFile(stderr,"    Offset: %.20g,%.20g\n",
+                (double) difference_image->page.x,(double)
+                difference_image->page.y);
+              (void) FormatLocaleFile(stderr,"    Similarity metric: %.*g\n",
+                GetMagickPrecision(),similarity_metric);
+            }
         }
       (void) ResetImagePage(difference_image,"0x0+0+0");
       if (difference_image->next != (Image *) NULL)
@@ -1455,14 +1477,7 @@ WandExport MagickBooleanType CompareImageCommand(ImageInfo *image_info,
     }
   DestroyCompare();
   image_info=restore_info;
-  if ((metric == NormalizedCrossCorrelationErrorMetric) ||
-      (metric == UndefinedErrorMetric))
-    {
-      if (fabs(distortion-1.0) > CompareEpsilon)
-        (void) SetImageOption(image_info,"compare:dissimilar","true");
-    }
-  else
-    if (fabs(distortion) > CompareEpsilon)
-      (void) SetImageOption(image_info,"compare:dissimilar","true");
+  if (similar == MagickFalse)
+    (void) SetImageOption(image_info,"compare:dissimilar","true");
   return(status != 0 ? MagickTrue : MagickFalse);
 }
